import type { Modifier } from '@popperjs/core';
import Component from '@ember/component';
/**
 * The UiContextElement component serves as the base class for tooltips, popovers, menus,
 * and more. Anything that needs to be dynamically positioned relative to some other "anchor"
 * element begins here.
 *
 * @class UiContextualElement
 */
export default class UiContextualElement extends Component {
    /**
     * The role attribute of the element.
     *
     * @argument ariaRole
     * @type {string}
     * @default "tooltip"
     * @public
     */
    ariaRole: string;
    /**
     * Placement of the element, relative to its reference element.
     * One of ['top', 'right', 'bottom', 'left'].
     *
     * @argument placement
     * @type {string}
     * @default "top"
     * @public
     */
    placement: string;
    /**
     * Whether or not the element is visible and positioned.
     *
     * @argument isHidden
     * @type {boolean}
     * @default false
     * @public
     */
    isHidden: boolean;
    /**
     * @argument fade
     * @type {boolean}
     * @default true
     * @public
     */
    fade: boolean;
    /**
     * @argument showContent
     * @type {boolean}
     * @default false
     * @public
     */
    showContent: boolean;
    /**
     * If true, the element will be displayed and positioned within its parent element. When
     * false, the element will be moved to a higher position in the DOM. Use this to mitigate
     * z-index and other layout issues.
     *
     * @argument renderInPlace
     * @type {boolean}
     * @default false
     * @public
     */
    renderInPlace: boolean;
    /**
     * If true, the element will change its position automatically based on available space.
     *
     * @argument autoPlacement
     * @type {boolean}
     * @default true
     * @public
     */
    autoPlacement: boolean;
    /**
     * Virtual "padding" for the PopperJS preventOverflow modifier.
     * https://popper.js.org/docs/v2/modifiers/prevent-overflow/#padding
     *
     * @argument viewportPadding
     * @type {number|{ top?: number, left?: number, right?: number, bottom?: number }}
     * @default 0
     * @public
     */
    viewportPadding: number;
    /**
     * An offset, in pixels, for the PopperJS offset modifier.
     * https://popper.js.org/docs/v2/modifiers/offset/#distance
     *
     * @argument distance
     * @type {number}
     * @default 0
     * @public
     */
    distance: number;
    /**
     * The "reference" that this element will be positioned relative to.
     *
     * @argument popperTarget
     * @type {HTMLElement}
     * @default undefined
     * @public
     * @required
     */
    popperTarget?: HTMLElement;
    /**
     * Describes the area that the element will be checked for overflow relative to.
     * https://popper.js.org/docs/v2/modifiers/prevent-overflow/#boundary
     *
     * @argument viewportElement
     * @type {HTMLElement|"clippingParents"}
     * @default undefined
     * @public
     * @required
     */
    viewportElement?: HTMLElement | 'clippingParents';
    /**
     * If `renderInPlace` is false, then this will need to be supplied as the
     * container that this element will be rendered into.
     *
     * @argument destinationElement
     * @type {HTMLElement}
     * @default undefined
     * @public
     */
    destinationElement?: HTMLElement;
    /**
     * @argument testId
     * @type {string}
     * @default undefined
     * @public
     */
    testId?: string;
    /**
     * CSS class names for the element.
     *
     * @argument popperClassNames
     * @type {string | string[]}
     * @default undefined
     * @public
     */
    get popperClassNames(): undefined | string | string[];
    /**
     * A unique id for the element. One will be generated if not provided.
     *
     * @argument id
     * @type {string}
     * @default undefined
     * @public
     */
    id?: string;
    /**
     * An optional value for the CSS max-width property of the inner block of this
     * element.
     *
     * @argument maxWidth
     * @type {string | number}
     * @default undefined
     * @public
     */
    maxWidth?: string | number;
    /**
     * If true, a directional caret will be shown, pointing to the target element.
     *
     * @argument showArrow
     * @type {boolean}
     * @default true
     * @public
     */
    showArrow: boolean;
    /**
     * CSS class name for the arrow caret.
     *
     * @argument arrowClassName
     * @type {string}
     * @default "tooltip-arrow"
     * @public
     */
    arrowClassName: string;
    /**
     * CSS class name for the inner block content of this element.
     *
     * @argument innerClassName
     * @type {string}
     * @default "tooltip-inner"
     * @public
     */
    innerClassName: string;
    /**
     * The real placement of the element. If `autoPlacement` is enabled then
     * this could differ from `placement`.
     *
     * @argument actualPlacement
     * @type {string}
     * @public
     * @readonly
     */
    readonly actualPlacement: string;
    /**
     * The `placement` property allows for either PopperJS or Bootstrap
     * standard layout position strings to be provided. This readonly
     * value will always be the PopperJS translated version.
     *
     * @argument popperPlacement
     * @type {string}
     * @public
     * @readonly
     */
    get popperPlacement(): string;
    /**
     * Concatenated CSS class names for the element.
     *
     * @argument popperClasses
     * @type {string}
     * @public
     * @readonly
     */
    get popperClasses(): string;
    /**
     * An array of PopperJS modifier descriptions created from the configuration
     * properties of this component.
     *
     * @argument popperModifiers
     * @type {Modifier[]}
     * @public
     * @readonly
     */
    get popperModifiers(): Modifier<unknown, Record<string, unknown>>[];
    /**
     * @argument innerElementStyles
     * @type {string | undefined}
     * @public
     * @readonly
     */
    get innerElementStyles(): import("@ember/template/-private/handlebars").SafeString | undefined;
    /**
     * @method getPopperModifiers
     * @return {Modifier[]}
     * @protected
     */
    protected getPopperModifiers(): Modifier<unknown, Record<string, unknown>>[];
    /**
     * PopperJS uses slightly different terminology than Bootstrap.
     *
     * @method updatePlacement
     * @param {unknown} data
     * @protected
     */
    protected updatePlacement(data?: {
        state?: {
            placement: string;
        };
    }): void;
}
